/* My first attempt at making a piece of software for the Sega Master System
   A simple game of Battleships
   
   Made by: Jens Kristian Krarup Keller
   
   2011

   ########################################################################*/

;==============================================================
; WLA-DX banking setup
;==============================================================
.memorymap
defaultslot 0
slotsize $8000
slot 0 $0000
.endme

.rombankmap
bankstotal 1
banksize $8000
banks 1
.endro

;==============================================================
; Defines
;==============================================================
.define VDPControl $BF
.define VDPData $BE
.define VRAMWrite $4000
.define CRAMWrite $c000

.define Joypad $DC
.define PSGSetup $7F

.define StartPlayerX 96
.define StartPlayerY 63
.define MaxPlayerX 152
.define MaxPlayerY 119

.define StartEnemyX 160
.define StartEnemyY 95
.define MaxEnemyX 216
.define MaxEnemyY 151

;==============================================================
; Enumerations (variables)
;==============================================================
.ENUM $C000
      ship_type db
      ship_pos_x db
      ship_pos_y db
      ship_length db
      ship_tile_nr db
      sprite_tile_nr db
      rotated db

      board_counter db
      player_board dsb 64
      enemy_board dsb 64
      
      crosshair_x db
      crosshair_y db
      
      enemy_air db
      enemy_bat db
      enemy_cru db
      enemy_des_1 db
      enemy_des_2 db
      enemy_sub_1 db
      enemy_sub_2 db
      enemy_ships db

      player_air db
      player_bat db
      player_cru db
      player_des_1 db
      player_des_2 db
      player_sub_1 db
      player_sub_2 db
      player_ships db
      
      sound_counter db
.ENDE

;==============================================================
; SDSC tag and SMS rom header
;==============================================================
.sdsctag 1.0,"Battleships","First attempt at making a game","Jens Kristian Krarup Keller"

.bank 0 slot 0
.org $0000
;==============================================================
; Boot section
;==============================================================
    di              ; disable interrupts
    im 1            ; Interrupt mode 1
    jp Main         ; jump to main program

.org $0038
;==============================================================
; Vblank/Hblank interrupt handler
;==============================================================
    push af
         in a,(VDPControl)      ;satisfy the IRQ

         call CountBoard

         ;Increment sound counter
         ld a, (sound_counter)
         inc a
         ld (sound_counter), a

    pop af
    ei   ;Enable new interrupt
    ret

.org $0066
;==============================================================
; Pause button handler
;==============================================================
    ; Do nothing
    retn

;==============================================================
; Main program
;==============================================================
Main:
    ld sp, $DFF0             ;Set stack

    ;==============================================================
    ; Set up VDP registers
    ;==============================================================
    ld hl,VDPInitData
    ld b,VDPInitDataEnd-VDPInitData
    ld c,VDPControl
    otir

    ;==============================================================
    ; Clear VRAM
    ;==============================================================
    ; 1. Set VRAM write address to $0000
    ld hl,$0000 | VRAMWrite
    call SetVDPAddress

    ; 2. Output 16KB of zeroes
    ld bc,$4000     ; Counter for 16KB of VRAM
-:  xor a
    out (VDPData),a ; Output to VRAM address, which is auto-incremented after each write
    dec bc
    ld a,b
    or c
    jr nz,-

    ;==============================================================
    ; Load palette
    ;==============================================================
    ; 1. Set VRAM write address to CRAM (palette) address 0
    ld hl,$0000 | CRAMWrite
    call SetVDPAddress

    ; 2. Output colour data
    ld hl,PaletteData
    ld bc,PaletteDataEnd-PaletteData
    call CopyToVDP

    ;==============================================================
    ; Load tiles
    ;==============================================================
    ; 1. Set VRAM write address to tile index 0
    ld hl,$0000 | VRAMWrite
    call SetVDPAddress

    ; 2. Output tile data
    ld hl,TileData              ; Location of tile data
    ld bc,TileDataEnd-TileData  ; Counter for number of bytes to write
    call CopyToVDP

    ;==============================================================
    ; Write first background to name table
    ;==============================================================
    ; 1. Set VRAM write address to tilemap index 0
    ld hl,$3800 | VRAMWrite
    call SetVDPAddress

    ; 2. Output background data
    ld hl,FirstBackgroundData
    ld bc,FirstBackgroundDataEnd-FirstBackgroundData  ; Counter for number of bytes to write
    call CopyToVDP

    ;==============================================================
    ; Clear the boards
    ;==============================================================
    ; Clear boards
    ld hl, player_board
    ld b, 64
-:  ld (hl), 0
    inc hl
    djnz -

    ld hl, enemy_board
    ld b, 64
-:  ld (hl), 0
    inc hl
    djnz -
    
    ; Clear counter
    xor a
    ld (board_counter), a
    
    ;==============================================================
    ; Give life to the ships and players
    ;==============================================================
    ld a, 5
    ld (enemy_air), a
    ld (player_air), a
    dec a
    ld (enemy_bat), a
    ld (player_bat), a
    dec a
    ld (enemy_cru), a
    ld (player_cru), a
    dec a
    ld (enemy_des_1), a
    ld (enemy_des_2), a
    ld (player_des_1), a
    ld (player_des_2), a
    dec a
    ld (enemy_sub_1), a
    ld (enemy_sub_2), a
    ld (player_sub_1), a
    ld (player_sub_2), a
    ld a, 7
    ld (enemy_ships), a
    ld (player_ships), a

    ;==============================================================
    ; Set sound
    ;==============================================================

    ; Channel 0 frequency
    ld a, $80
    out (PSGSetup), a
    ld a, $0E
    out (PSGSetup), a

    ; Channel 1 frequency
    ld a, $A0
    out (PSGSetup), a
    ld a, $20
    out (PSGSetup), a

    ; Channel 2 frequency
    ld a, $C0
    out (PSGSetup), a
    ld a, $11
    out (PSGSetup), a

    ; Noise
    ld a, $E7
    out (PSGSetup), a

    ;==============================================================
    ; Make an aircraft carrier
    ;==============================================================
    ;Set the variables
    ld a, %01000001          ;Aircraft carrier
    ld (ship_type), a
    ld a, 5                  ;5 tiles long
    ld (ship_length), a
    ld a, $0D                ;'A' tile
    ld (ship_tile_nr), a
    ld a, $1D                ;Inverted 'A' tile with no border
    ld (sprite_tile_nr), a
    xor a                    ;Don't rotate the ship (just to have an initial value)
    ld (rotated), a

    ;==============================================================
    ; Turn screen on
    ;==============================================================
    ld a, %11100000
    ;       |||| |`- Zoomed sprites -> 16x16 pixels
    ;       |||| `-- Doubled sprites -> 2 tiles per sprite, 8x16
    ;       |||`---- 30 row/240 line mode
    ;       ||`----- 28 row/224 line mode
    ;       |`------ VBlank interrupts
    ;       `------- Enable display
    out (VDPControl),a
    ld a, $81
    out (VDPControl),a

    ei                       ;Enable interrupts

    call PlacePlayer         ;Place the aircraft carrier
    call PlaceEnemy

    ;==============================================================
    ; Make a battleship
    ;==============================================================
    ;Set the variables
    ld a, %01000010          ;Battleship
    ld (ship_type), a
    ld a, 4                  ;4 tiles long
    ld (ship_length), a
    ld a, $0E                ;'B' tile
    ld (ship_tile_nr), a
    ld a, $1E                ;Inverted 'B' tile with no border
    ld (sprite_tile_nr), a
    xor a                    ;Don't rotate the ship (just to have an initial value)
    ld (rotated), a

    call PlacePlayer         ;Place the battleship
    call PlaceEnemy

    ;==============================================================
    ; Make a cruiser
    ;==============================================================
    ;Set the variables
    ld a, %01000011          ;Cruiser
    ld (ship_type), a
    ld a, 3                  ;3 tiles long
    ld (ship_length), a
    ld a, $0F                ;'C' tile
    ld (ship_tile_nr), a
    ld a, $1F                ;Inverted 'C' tile with no border
    ld (sprite_tile_nr), a
    xor a                    ;Don't rotate the ship (just to have an initial value)
    ld (rotated), a

    call PlacePlayer         ;Place the cruiser
    call PlaceEnemy

    ;==============================================================
    ; Make two destroyers
    ;==============================================================
    ;Set the variables
    ld a, %01000100          ;First destroyer
    ld (ship_type), a
    ld a, 2                  ;2 tiles long
    ld (ship_length), a
    ld a, $10                ;'D' tile
    ld (ship_tile_nr), a
    ld a, $20                ;Inverted 'D' tile with no border
    ld (sprite_tile_nr), a
    xor a                    ;Don't rotate the ship (just to have an initial value)
    ld (rotated), a

    call PlacePlayer         ;Place the first destroyer
    call PlaceEnemy

    ld a, %01000101          ;Second Destroyer
    ld (ship_type), a
    xor a                    ;Don't rotate the ship (just to have an initial value)
    ld (rotated), a

    call PlacePlayer         ;Place the second destroyer
    call PlaceEnemy

    ;==============================================================
    ; Make two submarines
    ;==============================================================
    ;Set the variables
    ld a, %01000110          ;First submarine
    ld (ship_type), a
    ld a, 1                  ;1 tile
    ld (ship_length), a
    ld a, $15                ;'S' tile
    ld (ship_tile_nr), a
    ld a, $21                ;Inverted 'S' tile with no border
    ld (sprite_tile_nr), a

    call PlacePlayer         ;Place the first submarine
    call PlaceEnemy

    ld a, %01000111          ;Second submarine
    ld (ship_type), a

    call PlacePlayer         ;Place the second submarine
    call PlaceEnemy
    
    ;==============================================================
    ; Write second background to name table
    ;==============================================================
    ; 1. Set VRAM write address to tilemap index 0
    ld hl,$3800 | VRAMWrite
    call SetVDPAddress

    ; 2. Output background data
    ld hl,SecondBackgroundData
    ld bc,SecondBackgroundDataEnd-SecondBackgroundData  ; Counter for number of bytes to write
    call CopyToVDP
    
    ;==============================================================
    ; Write player_board information to name table
    ;==============================================================
    ; 1. Set VRAM write address to start of player board on tilemap
    ld hl, $3908 | VRAMWrite
    call SetVDPAddress
    
    ; 2. Output first line of data from the player board
    
    ;Address of the first byte of the player board
    ld de, player_board
    call PrintBoardLine

    ; 3. Output the last 7 lines
    ld b, 7
-:  push bc
         ld bc, 64       ;Jump 32 tiles
         add hl, bc
         call SetVDPAddress
         call PrintBoardLine
    pop bc
    djnz -

    ;==============================================================
    ; Place crosshair at start position
    ;==============================================================
    ld a, StartEnemyX     ;160 pixels = 20 tiles in
    ld (crosshair_x), a
    ld a, StartEnemyY     ;95 + 1 pixels = 12 tiles down
    ld (crosshair_y), a

    ;==============================================================
    ; Main loop (the actual game)
    ;==============================================================
MainLoop:
    call PlayerStrike    ;The player strikes
    call EnemyStrike     ;The CPU strikes
    jp MainLoop

;==============================================================
; Functions
;==============================================================
; PlaceShip:
.include "place_ship.asm"

; PlaceEnemy:
.include "place_enemy.asm"

; PlayerStrike:
.include "player_strike.asm"

; EnemyStrike:
.include "enemy_strike.asm"

    ;==============================================================
    ; Helper functions
    ;==============================================================

    SetVDPAddress:
    ; Sets the VDP address
    ; Parameters: hl = address
    ; Affects: a
    ld a,l
    out (VDPControl),a
    ld a,h
    out (VDPControl),a
    ret

    CopyToVDP:
    ; Copies data to the VDP
    ; Parameters: hl = data address, bc = data length
    ; Affects: a, hl, bc
-:  ld a,(hl)    ; Get data byte
    out (VDPData),a
    inc hl       ; Point to next byte
    dec bc
    ld a,b
    or c
    jr nz,-
    ret

    StrikeSound:
    push af
         ld a, $F0
         out (PSGSetup), a
         xor a
         ld (sound_counter), a
-:       ld a,(sound_counter)
         cp 30
         jr nz,-
         ld a, $FF
         out (PSGSetup), a
    pop af
    ret
    
    MissSound:
    ;push af
         ld a, $B5
         out (PSGSetup), a
         xor a
         ld (sound_counter), a
-:       ld a,(sound_counter)
         cp 15
         jr nz,-
         ld a, $BF
         out (PSGSetup), a
    ;pop af
    ret
    
    DenySound:
    push af
         ld a, $95
         out (PSGSetup), a
         xor a
         ld (sound_counter), a
-:       ld a,(sound_counter)
         cp 15
         jr nz,-
         ld a, $9F
         out (PSGSetup), a
    pop af
    ret

    WaitAMoment:
    push af
         xor a
         ld (board_counter), a
-:       ld a,(board_counter)
         cp 50
         jr nz,-
    pop af
    ret
    
    DrawHole:
    ld a, $33
    out (VDPData), a
    xor a
    out (VDPData), a
    ret

    CountBoard:
    ; Count variable board_counter from 0 to 63
    ld a, (board_counter)
    inc a
    cp 64
    jr nz,+
    xor a
+:  ld (board_counter), a
    ret
    
    PrintBoardLine:
    ld b, 8              ;Process 8 tiles
-:  ld a, (de)           ;Load a byte of data
    and %00111111        ;Remove status bits

    ;Compare with ship types and output corresponding tiles
    cp 1
    jr nz,+
    ld a, $0D
    out (VDPData), a
    jp ++

+:  cp 2
    jr nz,+
    ld a, $0E
    out (VDPData), a
    jp ++

+:  cp 3
    jr nz,+
    ld a, $0F
    out (VDPData), a
    jp ++
    
+:  cp 4
    jr nz,+
    ld a, $10
    out (VDPData), a
    jp ++
    
+:  cp 5
    jr nz,+
    ld a, $10
    out (VDPData), a
    jp ++

+:  cp 6
    jr nz,+
    ld a, $15
    out (VDPData), a
    jp ++
    
+:  cp 7
    jr nz,+
    ld a, $15
    out (VDPData), a
    jp ++
    
+:  ld a, $04
    out (VDPData), a

++: xor a
    out (VDPData), a
    inc de
    djnz -
    ret

;==============================================================
; Data
;==============================================================
; VDP initialisation data
VDPInitData:
.db $04,$80,$00,$81,$FF,$82,$FF,$83,$FF,$84,$FF,$85,$00,$86,$00,$87,$00,$88,$00,$89,$FF,$8A
VDPInitDataEnd:

PaletteData:
; Background colors
.db $30,$2F,$03,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00
; Sprite colors
.db $30,$03,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00
PaletteDataEnd:

;Tiles:
.include "tiles.asm"

;Backgrounds:
.include "background.asm"